/****************************************************************************/
/*      Everything You Always Wanted to Know about Data Partitioning        */
/*                         but Were AfraId to Ask                           */
/*                            PASS Summit 2017                              */
/*                                                                          */
/*                  Written by Dmitri V. Korotkevitch                       */
/*                      http://aboutsqlserver.com                           */
/*                        dk@aboutsqlserver.com                             */
/****************************************************************************/
/*                               Initial Setup                              */
/*                    Populating Non-Partitioned DB with Data               */
/****************************************************************************/

/****************************************************************************/
/*  This script generates 10+ GB database and introduce heavy I/O activity  */
/****************************************************************************/


set xact_abort on
set nocount on
go

use NonPartitionedDB
go

if exists(select * from sys.tables t join sys.schemas s on t.schema_Id = s.schema_Id where s.name = 'dbo' and t.name='Orders') drop table dbo.Orders;


create table dbo.Orders
(
	OrderDate datetime2(0) not null,
	OrderId int identity(1,1) not null,
	WarehouseId int not null,	
	CustomerId int not null, 
	ZipCode char(5) not null,
	Amount money not null,
	DeliveryInstructions varchar(1000) null
);

declare
	@FirstDate datetime2(0) = '2017-01-01'
	,@LastDate datetime2(0) = '2017-11-06'
	,@OrdersPerDay int = 100000
	,@MaxWarehouses int = 20
	,@Mins int
	,@StartDate datetime2(0)
	,@StopDate datetime2(0)
	,@I int = 1


while @FirstDate < @LastDate
begin
	select @StopDate = dateadd(day,-1,dateadd(month,1,@FirstDate))
	set @Mins = datediff(minute,@FirstDate,@StopDate)
	;with Days(ADay)
	as
	(
		select @FirstDate
		union all
		select dateadd(day,1,ADay) from Days where ADay < @StopDate
	)
	,Data(OrderDate, Num)
	as
	(
		select dateadd(second, n.Number * 24. * 60. * 60 / @OrdersPerDay, d.ADay), 
			row_number() over (order by (select null)) 
		from Days d cross join dbo.Numbers n
		where n.Number < @OrdersPerDay
	) 
	insert into dbo.Orders with (tablockx) 
		(OrderDate,WarehouseId,CustomerId,Amount,ZipCode,DeliveryInstructions)
		select OrderDate, Num % @MaxWarehouses + 1,Num % 200000, Num % 100 + .99 
			,convert(char(5),Num % 10000)
			,'Delivery instruction for order ' + convert(varchar(32),OrderDate,121)
		from Data
		where OrderDate <= @LastDate;

	checkpoint;
	waitfor delay '00:00:05.000';
	raiserror('%d is done',0,1,@I) with nowait;
	select @FirstDate = dateadd(month,1,@FirstDate), @I += 1
end
go

update top (99.9) percent dbo.Orders with (tablockx)
set OrderDate = dateadd(month,-1,OrderDate)
where WarehouseId = 10 and OrderDate >= '2017-11-01'
go

-- Making more values in CI Histogram
update o with (tablockx)
set OrderDate = dateadd(hour,t.Hours,dateadd(day,t.Days,o.OrderDate))
from dbo.Orders o cross apply
	(
		select 
			case datepart(dw,o.OrderDate)
				when 7 then -1 - o.OrderId % 2
				when 0 then 1 + o.OrderId % 2
				else 0
			end as Days
			,case 
				when datepart(hour,o.OrderDate) between 18 and 23 then -6
				when datepart(hour,o.OrderDate) between 0 and 6 then 8
				else 0
			end as Hours
	) t
where 
	o.OrderId % 100 < 80 and OrderDate > '2017-01-03'
go

/*
select count(*) from dbo.Orders;

select top 10 * from dbo.Orders order by OrderDate;

select top 10 * from dbo.Orders order by OrderDate desc;

select month(OrderDate), count(*)
from dbo.Orders
group by month(OrderDate)
order by 1;
*/

create unique clustered index IdX_Orders_OrderDate_OrderId
on dbo.Orders(OrderDate, OrderId);

create index IdX_Orders_WarehouseId_OrderDate 
on dbo.Orders(WarehouseId, OrderDate)
go

-- Let's shrink files to make DB smaller on disk. Do not run in production! :)
declare
    @MaxFileSizeMB int
    ,@SQL nvarchar(max)

-- Obtaining the current file size 	
select @MaxFileSizeMB = size / 128.0 - convert(int,fileproperty(name,'SpaceUsed')) / 128.0
from sys.database_files
where name = 'NonPartitionedDB';

set @SQL = N'dbcc shrinkfile(N''NonPartitionedDB'',' + convert(nvarchar(32),@MaxFileSizeMB) + N');';
exec sp_executesql @SQL;
go

checkpoint;
dbcc shrinkfile (N'NonPartitionedDB_log' , 16)
go

alter database NonPartitionedDB set recovery full
go