/****************************************************************************/
/*      Everything You Always Wanted to Know about Data Partitioning        */
/*                         but Were AfraId to Ask                           */
/*                            PASS Summit 2017                              */
/*                                                                          */
/*                  Written by Dmitri V. Korotkevitch                       */
/*                      http://aboutsqlserver.com                           */
/*                        dk@aboutsqlserver.com                             */
/****************************************************************************/
/*                               Initial Setup                              */
/*                   Populating Partitioned DB CCI with Data                */
/****************************************************************************/

set noexec off
set xact_abort on
set nocount on
go

use PartitionedDB
go

if not exists(select * from sys.databases where name = 'NonPartitionedDB')
begin
	raiserror('Please create NonPartitionedDB with the data first',16,1)
	set noexec on
end
go

if not exists((select * from NonPartitionedDB.sys.tables t join NonPartitionedDB.sys.schemas s on t.schema_Id = s.schema_Id where s.name = 'dbo' and t.name='Orders'))
begin
	raiserror('Please create PartitionedDB.Orders table first',16,1)
	set noexec on
end
go

if not exists((select * from sys.tables t join sys.schemas s on t.schema_Id = s.schema_Id where s.name = 'dbo' and t.name='Orders2017'))
begin
	raiserror('Please create PartitionedDB.Orders2017 table first',16,1)
	set noexec on
end

if exists(select * from sys.views v join sys.schemas s on v.schema_Id = s.schema_Id where s.name = 'dbo' and v.name='OrdersWithCCI') drop view dbo.OrdersWithCCI;
if exists(select * from sys.tables t join sys.schemas s on t.schema_Id = s.schema_Id where s.name = 'dbo' and t.name='Orders2017_CCI') drop table dbo.Orders2017_CCI;
if exists(select * from sys.partition_schemes where name = 'psCCIOrders2017') drop partition scheme psCCIOrders2017;
if exists(select * from sys.partition_functions where name = 'pfCCIOrders2017') drop partition function pfCCIOrders2017;
go

create partition function pfCCIOrders2017(datetime2(0))
as range right for values
('2017-02-01','2017-03-01','2017-04-01','2017-05-01'
,'2017-06-01','2017-07-01','2017-08-01','2017-09-01'
,'2017-10-01','2017-11-01','2017-12-01','2018-01-01');

create partition scheme psCCIOrders2017 
as partition pfCCIOrders2017
all to ([CCIOrders]);
go

create table dbo.Orders2017_CCI
(
	OrderDate datetime2(0) not null,
	OrderId int not null,
	WarehouseId int not null,	
	CustomerId int not null, 
	ZipCode char(5) not null,
	Amount money not null,
	constraint CHK_Orders_2017_CCI
	check (OrderDate >= '2017-01-01' and OrderDate < '2017-10-01')
) on psCCIOrders2017(OrderDate);

insert into dbo.Orders2017_CCI with (TABLOCKX)
		(OrderDate, OrderId, WarehouseId, CustomerId, ZipCode, Amount)
	select OrderDate, OrderId, WarehouseId, CustomerId, ZipCode, Amount
	from dbo.Orders2017 with (TABLOCK);
go

create clustered columnstore index CCI_Orders2017_CCI
on dbo.Orders2017_CCI
with 
(
	data_compression = columnstore_archive on partitions(1 to 6),
	data_compression = columnstore on partitions(7 to 13)
)
on psCCIOrders2017(OrderDate);
go

create view dbo.OrdersWithCCI(OrderDate, OrderId, WarehouseId, CustomerId
	,ZipCode, Amount, DeliveryInstructions)
as
	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_11

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_10

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, NULL as DeliveryInstructions
	from dbo.Orders2017_CCI
go

-- Let's shrink files to make DB smaller on disk. Do not run on production db :)
declare
    @MaxFileSizeMB int
    ,@SQL nvarchar(max)

-- Obtaining the current file size 	
select @MaxFileSizeMB = size / 128.0 - convert(int,fileproperty(name,'SpaceUsed')) / 128.0
from sys.database_files
where name = 'PartitionedDB_CurrentOrders';

set @SQL = N'dbcc shrinkfile(N''PartitionedDB_CurrentOrders'',' + convert(nvarchar(32),@MaxFileSizeMB) + N');';
exec sp_executesql @SQL;

select @MaxFileSizeMB = size / 128.0 - convert(int,fileproperty(name,'SpaceUsed')) / 128.0
from sys.database_files
where name = 'PartitionedDB_RecentOrders';

set @SQL = N'dbcc shrinkfile(N''PartitionedDB_RecentOrders'',' + convert(nvarchar(32),@MaxFileSizeMB) + N');';
exec sp_executesql @SQL;

select @MaxFileSizeMB = size / 128.0 - convert(int,fileproperty(name,'SpaceUsed')) / 128.0
from sys.database_files
where name = 'PartitionedDB_OldOrders';

set @SQL = N'dbcc shrinkfile(N''PartitionedDB_OldOrders'',' + convert(nvarchar(32),@MaxFileSizeMB) + N');';
exec sp_executesql @SQL;

select @MaxFileSizeMB = size / 128.0 - convert(int,fileproperty(name,'SpaceUsed')) / 128.0
from sys.database_files
where name = 'PartitionedDB_CCIOrders';

set @SQL = N'dbcc shrinkfile(N''PartitionedDB_CCIOrders'',' + convert(nvarchar(32),@MaxFileSizeMB) + N');';
exec sp_executesql @SQL;
go

checkpoint;
dbcc shrinkfile (N'PartitionedDB_log' , 16)
go


alter database PartitionedDB set recovery full
go