/****************************************************************************/
/*      Everything You Always Wanted to Know about Data Partitioning        */
/*                         but Were AfraId to Ask                           */
/*                            PASS Summit 2017                              */
/*                                                                          */
/*                  Written by Dmitri V. Korotkevitch                       */
/*                      http://aboutsqlserver.com                           */
/*                        dk@aboutsqlserver.com                             */
/****************************************************************************/
/*                              Moving Data                                 */
/*                           Moving Partition                               */
/****************************************************************************/

use PartitionedDb
go

if exists(select * from sys.tables t join sys.schemas s on t.schema_id = s.schema_id where t.name = 'Orders2017_09' and s.name = 'dbo') drop table dbo.Orders2017_09;
go

select 
	p.partition_number
	,fg.name as [Filegroup]
	,p.[Rows] 
	,convert(date,pv.value) as [Max Data Value] -- Belongs to next partition
	,sum(a.total_pages) as TotalPages 
	,sum(a.used_pages) as UsedPages 
	,sum(a.data_pages) as DataPages
	,sum(a.total_pages) * 8 as TotalSpaceKB 
	,sum(a.used_pages) * 8 as UsedSpaceKB 
	,sum(a.data_pages) * 8 as DataSpaceKB
from 
	sys.tables t with (nolock) join sys.indexes i with (nolock) on
		t.object_id = i.object_id
	join sys.partitions p with (nolock) on 
		i.object_id = p.object_id AND i.index_id = p.index_id
	join sys.allocation_units a with (nolock) on 
		p.partition_id = a.container_id
	join sys.filegroups fg with (nolock) on 
		a.data_space_id = fg.data_space_id
	join sys.partition_range_values pv with (nolock) on
		pv.boundary_id = p.partition_number
	join sys.partition_functions pf with (nolock) on 
		pv.function_id = pf.function_id
where
	t.object_id = object_id(N'dbo.Orders2017') and
	i.index_id = 1 and
	pf.name = 'pfOrders2017'
group by
	p.partition_number, p.[Rows], fg.name, pv.value
order by 
	p.partition_number;

create table dbo.Orders2017_09
(
	OrderDate datetime2(0) not null,
	OrderId int not null,
	WarehouseId int not null,	
	CustomerId int not null, 
	ZipCode char(5) not null,
	Amount money not null,
	constraint CHK_Orders_2017_09
	check (OrderDate >= '2017-09-01' and OrderDate < '2017-10-01')
) on [RecentOrders];

create unique clustered index IDX_Orders2017_09_OrderDate_OrderId
on dbo.Orders2017_09(OrderDate,OrderId)
with (data_compression=page)
on [RecentOrders];
go

create nonclustered index IDX_Orders2017_09_WarehouseId_OrderDate 
on dbo.Orders2017_09(WarehouseId, OrderDate)
include(ZipCode, Amount)
with (data_compression=page)
on [RecentOrders];

alter table dbo.Orders2017
switch partition 9 
to dbo.Orders2017_09;
go

alter view dbo.Orders(OrderDate, OrderId, WarehouseId, CustomerId
	,ZipCode, Amount, DeliveryInstructions)
as
	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_11

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_10

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, NULL as DeliveryInstructions
	from dbo.Orders2017_09

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, NULL as DeliveryInstructions
	from dbo.Orders2017
go

/* Things to consider:
1. 2017_09 Data modifications - needs to be handled in Data Access. Alternatively,
it is possible to create INSTEAD OF triggers on the view.

2. If partition is moved to another table - you need to change CHECK constraint in
the source table */

-- You need to rebuild to different partition scheme if table has LOB columns
create unique clustered index IDX_Orders2017_09_OrderDate_OrderId
on dbo.Orders2017_09(OrderDate,OrderId)
with (data_compression=page, online=on, drop_existing = on)
on [OldOrders];
go

create nonclustered index IDX_Orders2017_09_WarehouseId_OrderDate 
on dbo.Orders2017_09(WarehouseId, OrderDate)
include(ZipCode, Amount)
with (data_compression=page, online=on, drop_existing = on)
on [OldOrders];
go

alter partition scheme psOrders2017
next used [OldOrders]
go

begin tran
	alter partition function pfOrders2017()
	merge range ('2017-09-01');

	alter partition scheme psOrders2017
	next used [OldOrders];

	alter partition function pfOrders2017()
	split range ('2017-09-01');

	alter table dbo.Orders2017_09
	switch to dbo.Orders2017 
	partition 9; 
commit
go

alter view dbo.Orders(OrderDate, OrderId, WarehouseId, CustomerId
	,ZipCode, Amount, DeliveryInstructions)
as
	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_11

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_10

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, NULL as DeliveryInstructions
	from dbo.Orders2017
go


select 
	p.partition_number
	,fg.name as [Filegroup]
	,p.[Rows] 
	,convert(date,pv.value) as [Max Data Value] -- Belongs to next partition
	,sum(a.total_pages) as TotalPages 
	,sum(a.used_pages) as UsedPages 
	,sum(a.data_pages) as DataPages
	,sum(a.total_pages) * 8 as TotalSpaceKB 
	,sum(a.used_pages) * 8 as UsedSpaceKB 
	,sum(a.data_pages) * 8 as DataSpaceKB
from 
	sys.tables t with (nolock) join sys.indexes i with (nolock) on
		t.object_id = i.object_id
	join sys.partitions p with (nolock) on 
		i.object_id = p.object_id AND i.index_id = p.index_id
	join sys.allocation_units a with (nolock) on 
		p.partition_id = a.container_id
	join sys.filegroups fg with (nolock) on 
		a.data_space_id = fg.data_space_id
	join sys.partition_range_values pv with (nolock) on
		pv.boundary_id = p.partition_number
	join sys.partition_functions pf with (nolock) on 
		pv.function_id = pf.function_id
where
	t.object_id = object_id(N'dbo.Orders2017') and
	i.index_id = 1 and
	pf.name = 'pfOrders2017'
group by
	p.partition_number, p.[Rows], fg.name, pv.value
order by 
	p.partition_number;
go

