/****************************************************************************/
/*      Everything You Always Wanted to Know about Data Partitioning        */
/*                         but Were AfraId to Ask                           */
/*                            PASS Summit 2017                              */
/*                                                                          */
/*                  Written by Dmitri V. Korotkevitch                       */
/*                      http://aboutsqlserver.com                           */
/*                        dk@aboutsqlserver.com                             */
/****************************************************************************/
/*                              Moving Data                                 */
/*               Using Partitioned Tables and Views Together                */
/****************************************************************************/

use PartitionedDb
go

if exists(select * from sys.tables t join sys.schemas s on t.schema_Id = s.schema_Id where s.name = 'dbo' and t.name='Orders2017_12') drop table dbo.Orders2017_12;
go

-- 01. Creating Table for the next month
create table dbo.Orders2017_12
(
	OrderDate datetime2(0) not null,
	OrderId int identity(1,1) not null,
	WarehouseId int not null,	
	CustomerId int not null, 
	ZipCode char(5) not null,
	Amount money not null,
	DeliveryInstructions varchar(1000) null,

	constraint CHK_Orders_2017_12
	check (OrderDate >= '2017-12-01' and OrderDate < '2018-10-01')
) on [CurrentOrders];

create unique clustered index IDX_Orders2017_12_OrderDate_OrderId
on dbo.Orders2017_12(OrderDate,OrderId)
on [CurrentOrders]

create nonclustered index IDX_Orders2017_12_WarehouseId_OrderDate 
on dbo.Orders2017_12(WarehouseId, OrderDate)
on [CurrentOrders];
go

-- 02. Doing schema restructuring if needed
-- In this example we are removing DeliveryInstructionr from Orders2017_10

alter view dbo.Orders(OrderDate, OrderId, WarehouseId, CustomerId
	,ZipCode, Amount, DeliveryInstructions)
as
	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_11

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount,  NULL as DeliveryInstructions
	from dbo.Orders2017_10

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, NULL as DeliveryInstructions
	from dbo.Orders2017
go

alter table dbo.Orders2017_10 drop column DeliveryInstructions;
go

-- 02. Moving 2017-10 table to [RecentOrders] FG
-- This also removes DeliveryInstruction bytes from table rows
create unique clustered index IDX_Orders2017_10_OrderDate_OrderId
on dbo.Orders2017_10(OrderDate,OrderId)
with (data_compression=page, online=on, drop_existing = on)
on [RecentOrders];

create nonclustered index IDX_Orders2017_10_WarehouseId_OrderDate 
on dbo.Orders2017_10(WarehouseId, OrderDate)
include(ZipCode, Amount)
with (data_compression=page, online=on, drop_existing = on)
on [RecentOrders];
go

-- Switching partition and altering view
set xact_abort on
begin tran
	alter table dbo.Orders2017 drop constraint CHK_Orders_2017_01_09;

	alter table dbo.Orders2017_10
	switch to dbo.Orders2017 
	partition 10; 

	exec sp_executesql N'
alter view dbo.Orders(OrderDate, OrderId, WarehouseId, CustomerId
	,ZipCode, Amount, DeliveryInstructions)
as
	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_12

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, DeliveryInstructions
	from dbo.Orders2017_11

	union all

	select OrderDate, OrderId, WarehouseId, CustomerId
		,ZipCode, Amount, NULL as DeliveryInstructions
	from dbo.Orders2017';
commit
go


select 
	p.partition_number
	,fg.name as [Filegroup]
	,p.[Rows] 
	,convert(date,pv.value) as [Max Data Value] -- Belongs to next partition
	,sum(a.total_pages) as TotalPages 
	,sum(a.used_pages) as UsedPages 
	,sum(a.data_pages) as DataPages
	,sum(a.total_pages) * 8 as TotalSpaceKB 
	,sum(a.used_pages) * 8 as UsedSpaceKB 
	,sum(a.data_pages) * 8 as DataSpaceKB
from 
	sys.tables t with (nolock) join sys.indexes i with (nolock) on
		t.object_id = i.object_id
	join sys.partitions p with (nolock) on 
		i.object_id = p.object_id AND i.index_id = p.index_id
	join sys.allocation_units a with (nolock) on 
		p.partition_id = a.container_id
	join sys.filegroups fg with (nolock) on 
		a.data_space_id = fg.data_space_id
	join sys.partition_range_values pv with (nolock) on
		pv.boundary_id = p.partition_number
	join sys.partition_functions pf with (nolock) on 
		pv.function_id = pf.function_id
where
	t.object_id = object_id(N'dbo.Orders2017') and
	i.index_id = 1 and
	pf.name = 'pfOrders2017'
group by
	p.partition_number, p.[Rows], fg.name, pv.value
order by 
	p.partition_number;
go

