/****************************************************************************/
/*      Everything You Always Wanted to Know about Data Partitioning        */
/*                         but Were AfraId to Ask                           */
/*                            PASS Summit 2017                              */
/*                                                                          */
/*                  Written by Dmitri V. Korotkevitch                       */
/*                      http://aboutsqlserver.com                           */
/*                        dk@aboutsqlserver.com                             */
/****************************************************************************/
/*                        Going Beyond B-Tree                               */
/*                            Data Movement                                 */
/****************************************************************************/

set nocount on
set xact_abort on
go

-- In SQL Server 2014 you can replace triggers with the code on data access tier

use InMemoryOLTP2016
go

drop trigger if exists dbo.trgAfterUpdate;
drop trigger if exists dbo.trgAfterDelete;
drop table if exists dbo.PaymentsUpdateQueue;
drop table if exists dbo.PaymentsDeleteQueue;
drop table if exists dbo.Payments2017_10_Tmp;
go

create table dbo.PaymentsUpdateQueue
(
	ID int not null identity(1,1)
		constraint PK_PaymentsUpdateQueue
		primary key nonclustered hash
		with (bucket_count=262144),
	PaymentId uniqueidentifier not null, 
)
with (memory_optimized=on, durability=schema_and_data)
go


create table dbo.PaymentsDeleteQueue
(
	ID int not null identity(1,1)
		constraint PK_PaymentsDeleteQueue
		primary key nonclustered hash
		with (bucket_count=262144),
	PaymentId uniqueidentifier not null
)
with (memory_optimized=on, durability=schema_and_data)
go

create trigger trgAfterUpdate on dbo.Payments2017_10
with native_compilation, schemabinding
after update
as
begin atomic with
(
    transaction isolation level = snapshot
    ,language = N'English'
)
	insert into dbo.PaymentsUpdateQueue(PaymentId)
		select PaymentId 
		from inserted
end
go

create trigger trgAfterDelete on dbo.Payments2017_10
with native_compilation, schemabinding
after delete
as
begin atomic with
(
    transaction isolation level = snapshot
    ,language = N'English'
)
	insert into dbo.PaymentsDeleteQueue(PaymentId)
		select PaymentId 
		from deleted
end
go

create table dbo.Payments2017_10_Tmp
(
	PaymentDate datetime2(0) not null,
	PaymentId uniqueidentifier not null, 
	CustomerId int not null,
	Amount money not null,
	Status tinyint not null,

	check (PaymentDate >= '2017-10-01' and PaymentDate < '2017-11-01')
)
on [FG2017]
go

create unique clustered index IDX_Payments2017_10_Tmp_PaymentDate_PaymentId
on dbo.Payments2017_10_Tmp(PaymentDate, PaymentId)
with (data_compression=row)
on [FG2017]
go

create nonclustered index IDX_Payments2017_10_Tmp_CustomerId
on  dbo.Payments2017_10_Tmp(CustomerId)
with (data_compression=row)
on [FG2017]
go

create nonclustered index IDX_Payments2017_10_Tmp_PaymentId
on  dbo.Payments2017_10_Tmp(PaymentId)
with (data_compression=row)
on [FG2017]
go

-- Step 1: Move data from the main table
insert into dbo.Payments2017_10_Tmp(PaymentDate, PaymentId, CustomerId, Amount, Status)
	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2017_10 with (snapshot)
go

-- Step 2: Update/Delete
select count(*) as [In Update Queue] from dbo.PaymentsUpdateQueue with (snapshot)
select count(*) as [In Delete Queue] from dbo.PaymentsDeleteQueue with (snapshot)
go

declare
	@MaxUpdateId int
	,@MaxDeleteId int

select @MaxUpdateId = max(ID) from dbo.PaymentsUpdateQueue with (snapshot);
select @MaxDeleteId = max(ID) from dbo.PaymentsDeleteQueue with (snapshot);

begin tran
	 if @MaxUpdateId is not null
	begin
		update t
		set t.Amount = s.Amount, t.Status = s.Status
		from 
			dbo.PaymentsUpdateQueue q with (snapshot) join dbo.Payments2017_10 s with (snapshot) on
				q.PaymentId = s.PaymentId 
			join dbo.Payments2017_10_Tmp t on
				t.PaymentId = s.PaymentId
		where
			q.ID <= @MaxUpdateId;
		
		delete from dbo.PaymentsUpdateQueue with (snapshot)
		where ID <= @MaxUpdateId;
	end;

	if @MaxDeleteId is not null
	begin
		delete from t
		from 
			dbo.PaymentsDeleteQueue q with (snapshot) join dbo.Payments2017_10_Tmp t on
				t.PaymentId = q.PaymentId
		where
			q.ID <= @MaxDeleteId;
		
		delete from dbo.PaymentsDeleteQueue with (snapshot)
		where ID <= @MaxDeleteId;
	end

	select count(*) as [In Update Queue] from dbo.PaymentsUpdateQueue with (snapshot)
	select count(*) as [In Delete Queue] from dbo.PaymentsDeleteQueue with (snapshot)
commit
go

-- Step 3: Replace
-- At the final phase, you need to do the following:
-- 1. Prevent the client from updating/deleting 2017_10 payments
-- 2. Copy "tail" from Queue tables 
-- 3. Swotch Payments2017_11 as partition of Payments2017 table
-- 4. Recreate 2017_11 SPs pointing to Payments2017 table
-- 5. Drop Payments2017_11 table
-- 6. Create 2017_12 table and SPs
-- 7. Alter the view.

-- Those are very quick actions; however, they may lead to exceptions if client tries to insert
-- 2017_10 rows in the middle of the process. The best way to address it is TRY/CATCH logic. 
-- You can also switch 2017_10 SPs to be interop ones and alter them in the beginning of transaction
-- to obtain Sch-M lock and block client from calling them.

set xact_abort on
go

drop proc dbo.InsertPayments2017_10;
drop proc dbo.UpdatePayments2017_10;
drop proc dbo.DeletePayments2017_10;
go

begin tran 
go

create proc dbo.UpdatePayments2017_10
(
	@PaymentDate datetime2(0) 
	,@PaymentId uniqueidentifier
	,@Amount money
	,@Status tinyint
)
as
	update dbo.Payments2017
	set Amount = @Amount, @Status = @Status
	where PaymentId = @PaymentId and @PaymentDate = @PaymentDate
go

create proc dbo.DeletePayments2017_10
(
	@PaymentDate datetime2(0) 
	,@PaymentId uniqueidentifier
)
as
	delete from dbo.Payments2017
	where PaymentId = @PaymentId and @PaymentDate = @PaymentDate
go

update t
set t.Amount = s.Amount, t.Status = s.Status
from 
	dbo.PaymentsUpdateQueue q with (snapshot) join dbo.Payments2017_10 s with (snapshot) on
		q.PaymentId = s.PaymentId 
	join dbo.Payments2017_10_Tmp t on
		t.PaymentId = s.PaymentId;

delete from t
from 
	dbo.PaymentsDeleteQueue q with (snapshot) join dbo.Payments2017_10_Tmp t on
		t.PaymentId = q.PaymentId;
go

alter table dbo.Payments2017
	drop constraint CHK_Payment2017_01_09;
go

alter table dbo.Payments2017_10_Tmp switch to dbo.Payments2017 partition 10;
go

alter view dbo.Payments(PaymentDate, PaymentId, CustomerId, Amount, Status)
as
	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2017_11
	
	union all

	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2017

	union all

	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2016
go

commit
go

drop table dbo.Payments2017_10
go

-- Create 2017_12 tables/code and ALTER VIEW again here (outside of tran to minimize downtime)

select max(PaymentDate) from dbo.Payments2017
go
